/*

sudosh - sudo shell that supports input and output logging to syslog

Copyright 2004 and $Date: 2004/10/10 03:57:00 $
                Douglas Richard Hanks Jr.

Licensed under the Open Software License version 2.0

This program is free software; you can redistribute it
and/or modify it under the terms of the Open Software
License, version 2.0 by Lauwrence E. Rosen.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the Open Software License for details.

*/

#include <errno.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <time.h>
#include <syslog.h>
#include <string.h>
#include <pwd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <regex.h>
#include <fcntl.h>
#include <stropts.h>
#include <sys/select.h>
#include <libgen.h>
#include <sys/param.h>

#include "config.h"

#define BEL 0x07
#define BS 0x08
#define TAB 0x09
#define LF 0x0A
#define VT 0x0B
#define FF 0x0C
#define CR 0x0D
#define SO 0x0E
#define SI 0x0F
#define CAN 0x18
#define SUB 0x1A
#define ESC 0x1B
#define DEL 0x7F

#define SS2 0x8E
#define SS3 0x8F
#define DCS 0x90
#define CSI 0x9B
#define ST 0x9C
#define OSC 0x9D
#define PM 0x9E
#define APC 0x9F

#define MAXBUF         8192

enum { Normal, Esc, Csi, Dcs, DcsString, DropOne, DecSet } state = Normal;

/* Borrowed from James Carlonson from groups.google.com */

void filter(char *str)
{
	int len = strlen(str);
	char *ptr = str;
	char *cstr = (char *)malloc(len + 1);
	char *cptr = cstr;
	int chr;

	while((chr = *ptr++) && (len-- > 0))
	{
		chr &= 0xFF;
		if(state == DropOne)
		{
			state = Normal;
			continue;
		}

		if(chr == DEL || chr == BS)
		{
			if(!*(cptr-1) == 0)
				*cptr--;
			continue;
		}		

		if(state == Esc && chr >= 0x40 && chr <= 0x5F)
		{
			state = Normal;
			chr += 0x40;
		}

		switch (chr)
		{
			case CAN:
			case SUB:
				state = Normal;
				break;
			case ESC:
				state = Esc;
				break;
			case CSI:
				state = Csi;
				break;
			case DCS:
			case OSC:
			case PM:
			case APC:
				state = Dcs;
				break;
			default:
				if((chr & 0x6F) < 0x20)
				{
					switch (chr)
					{
						case BEL:
						case BS:
						case TAB:
						case LF:
						case VT:
						case FF:
						case CR:
							*cptr++ = chr;
							break;
					}
					break;
				}
				switch (state)
				{
					case Normal:
						*cptr++ = chr;
						break;
					case Esc:
						state = Normal;
						switch (chr)
						{
							case 'c':
							case '7':
							case '8':
							case '=':
							case '>':
							case '~':
							case 'n':
							case '\123':
							case 'o':
							case '|':
								break;
							case '#':
							case ' ':
							case '(':
							case ')':
							case '*':
							case '+':
								state = DropOne;
							break;
						}
						break;
					case Csi:
					case Dcs:
						if(chr >= 0x40 && chr <= 0x7E)
						{
							if(state == Csi)
								state = Normal;
							else
								state = DcsString;
						}
						break;
					case DecSet:
						if(chr == 0x68)
							state = Normal;
						break;
					case DcsString:
						break;
					case DropOne:
						break;
				}
		}
	}

	*cptr = '\0';
	strcpy(str, cstr);
	free((void *)cstr);
}

/* Borrowed from Gerhard Lausser */

void write2syslog(const char *optr, size_t olen)
{
	static size_t rlen = 0;
	static char *rptr;
	char *eptr;
	char *tptr;
	char *lptr;
	static int rflag = 0;
	static int linecnt = 0;
	FILE *flog;
	char ff[MAXBUF];
	struct passwd *p = NULL;
        char *home = getenv("HOME");
        char *sudo_user = getlogin();

	if(home == NULL)
	{
		fprintf(stderr, "HOME variable is not set.  Stop.\n");
		exit(255);
	}

	if(sudo_user == NULL)
	{
		fprintf(stderr, "I have no idea who you are.  Stop.\n");
		exit(255);
	}

	p = getpwnam(sudo_user);

	if(p == NULL)
	{
		perror("getpwnam");
		exit(255);
	}

	snprintf(ff, (size_t) MAXBUF, "%s/.sudosh_input", p->pw_dir);

	if(olen == 0 || optr == NULL)
		return;

	if(rflag == 1 && *optr == '\n')
	{
		optr++;
		olen--;
	}

	rflag = 0;

	if(rptr == NULL)
		rptr = malloc(olen);
	else
		rptr = realloc(rptr, rlen + olen);

	memcpy(rptr + rlen, optr, olen);
	rlen += olen;
	tptr = rptr;
	lptr = memchr(tptr, '\r', rlen);

	while((lptr != NULL) && (tptr < (rptr + rlen)))
	{
		rflag = 1;
		eptr = malloc(lptr - tptr + 1);
		memcpy(eptr, tptr, lptr - tptr);
		*(eptr + (lptr - tptr)) = '\0';

		filter(eptr);

		flog = fopen(ff, "a");

		if(flog == NULL)
		{
			perror("fopen");
			exit(errno);
		}

		time_t now;
		int sec, min, hour, day, month, year;

		now = time(NULL);
		year = localtime(&now)->tm_year + 1900;
		month = localtime(&now)->tm_mon + 1;
		day = localtime(&now)->tm_mday;
		hour = localtime(&now)->tm_hour;
		min = localtime(&now)->tm_min;
		sec = localtime(&now)->tm_sec;

		fprintf(flog, "%02d/%02d/%04d %02d:%02d:%02d %03d: %s\n", day, month, year, hour, min, sec, linecnt, eptr);
		syslog(LOG_USER|LOG_NOTICE, "%03d: %s", linecnt++, eptr);
		fclose(flog);

		if(linecnt == 101)
			linecnt = 0;

		free(eptr);

		if((lptr == rptr + rlen - 1) && (*lptr == '\r'))
		{
			tptr = lptr + 1;
			lptr = NULL;
		}
		else if(lptr < rptr + rlen - 1)
		{
			tptr = lptr + 1;
			if((*tptr == '\n') && (tptr == (rptr + rlen - 1)))
			{
				lptr = NULL;
			}
			else if((*tptr == '\n') && (tptr < (rptr + rlen - 1)))
			{
				tptr++;
				lptr = memchr(tptr, '\r', rlen - (tptr - rptr));
			}
			else
				lptr = memchr(tptr, '\r', rlen - (tptr - rptr));
		}
		else
			lptr = NULL;
	  }

	while((tptr < (rptr + rlen)) && (*(tptr) == '\n'))
		tptr++;

	if((lptr == NULL) && (tptr < (rptr + rlen)))
	{
		memmove(rptr, tptr, ((rptr + rlen) - tptr));
		rlen = (rptr + rlen) - tptr;
		rptr = realloc(rptr, (size_t)rlen);
	} 
	else
	{
		free(rptr);
		rptr = NULL;
		rlen = 0;
	}
}
